ImageView subclass:#ImageEditView
	instanceVariableNames:'magnification gridMagnification selectColors imageReaderClass
		resourceClass resourceSelector editMode mouseKeyColorMode
		undoImage modified coordInfoBlock'
	classVariableNames:'Clipboard'
	poolDictionaries:''
	category:'Views-Misc'
!


!ImageEditView methodsFor:'accessing'!

coordInfoBlock: aBlock

    coordInfoBlock := aBlock
!

gridMagnification: aPoint

    gridMagnification := aPoint
!

imageReaderClass

    ^imageReaderClass
!

magnification

    ^magnification

!

magnification:aPoint

    magnification ~= aPoint
    ifTrue:
    [
        magnification := aPoint asPoint.
        self scrollToTopLeft.
        self contentsChanged.
        self invalidate.
    ].
!

resourceClass

    ^resourceClass
!

resourceClass: aClassOrSymbol

    resourceClass := aClassOrSymbol isClass ifTrue: [aClassOrSymbol name] ifFalse: [aClassOrSymbol asSymbol].

!

resourceMessage

    ^resourceClass, ' ', resourceSelector
!

resourceMessage: aMessage

    (aMessage isString and: [aMessage trimBlanks size > 0])
    ifTrue:
    [
        resourceClass := (aMessage readStream upTo: Character space) asSymbol.
        resourceSelector := (aMessage copy reverse readStream upTo: Character space) reverse asSymbol.
    ]
    ifFalse:
    [
        ^nil
    ].

   
!

resourceSelector

    ^resourceSelector
!

resourceSelector: aStringOrSymbol

    resourceSelector := aStringOrSymbol asSymbol
!

selectColor: anArrayTwoColors

    selectColors := anArrayTwoColors
!

selectedColor

   ^selectColors at: mouseKeyColorMode
!

selectedColor: aColor

    selectColors at: mouseKeyColorMode put: aColor
! !

!ImageEditView methodsFor:'drawing'!

redrawImageX:x y:y width:w height:h
    |ih iw dotW dotH minX maxX minY maxY color last lastY runW x0 xI yI maskColor|

    ih := image height.
    iw := image width.
    dotW := magnification x.
    dotH := magnification y.

    minX := x // dotW.
    minX >= iw ifTrue:[minX := iw - 1].
    minY := y // dotH.
    minY >= ih ifTrue:[minY := ih - 1].
    maxX := (x + w) // dotW + 1.
    maxX > iw ifTrue:[maxX := iw].
    maxY := (y + h) // dotH + 1.
    maxY > ih ifTrue:[maxY := ih].

    lastY := -1.

    x0 := minX.
    runW := 0.
    maskColor := false.
    image colorsFromX:minX y:minY toX:maxX-1 y:maxY-1 do:
    [:xx :yy :color|
        yy ~~ lastY
        ifTrue:
        [
            runW ~~ 0
            ifTrue:
            [
                |origin|
                origin := (x0 * dotW + margin)@(lastY * dotH + margin).
                self fillRectangle: (origin extent: (runW@dotH)).                    
                0 to: runW by: dotW do: [:xxx| self drawFrameAt: ((origin x + xxx) @(origin y))].
                maskColor ifTrue:
                [
                    self drawMaskPointAt: origin
                ].
                runW := 0.
            ]. 
            x0 := xx.
            lastY := yy.
        ]. 

        color ~~ last
        ifTrue:
        [
            runW ~~ 0
            ifTrue:
            [
                |origin|
                origin := (x0 * dotW + margin)@(yy * dotH + margin).
                self fillRectangle: (origin extent: (runW@dotH)).
                0 to: runW by: dotW do: [:xxx| self drawFrameAt: ((origin x + xxx)@(origin y))].
                maskColor ifTrue:
                [
                    self drawMaskPointAt: origin
                ].
                runW := 0.
            ].

            self paint: (last := color).
            image mask notNil ifTrue:
            [  
                maskColor := false.
                (image mask colorAt: xx@yy) = Color black ifTrue:
                [
                    self paint: (last := self viewBackground).
                    maskColor := true.
                ].
                last := nil.
            ].
            runW := 0.
            x0 := xx.
        ].  
        runW := runW + dotW
    ].
    runW ~~ 0 ifTrue:
    [
        |origin|
        origin := (x0 * dotW + margin)@(lastY * dotH + margin).
        self fillRectangle: (origin extent: runW@dotH).
        0 to: runW by: dotW do: [:xxx| self drawFrameAt: ((origin x + xxx) @origin y)].
        maskColor ifTrue:
        [
            self drawMaskPointAt: origin.
        ].
        runW := 0.
    ].
!

redrawX:x y:y width:w height:h
    |ih iw xI yI|

    image isNil ifTrue:[^self].

    self clippingRectangle: (x@y extent: w@h). 

    self redrawImageX:x y:y width:w height:h.

    "/ right of image ?
    adjust == #center ifTrue:
    [
        xI := (width - ih) // 2 - margin.
        yI := (height - iw) // 2 - margin.
    ]
    ifFalse:
    [
        xI := yI := margin
    ].
    (x + w - 1) > (xI + (magnification x * image width)) ifTrue:
    [
        self clearRectangleX:(xI + (magnification x * image width))
                           y:y
                       width:(x + w - (magnification x * image width) - xI)
                      height:h
    ].
    (y + h - 1) > (yI + (magnification y * image height)) ifTrue:
    [
        self clearRectangleX:margin
                           y:(yI + (magnification y * image height))
                       width:w
                      height:(y + h - (magnification y * image height) - yI)  
    ].
    self drawFrame.
    self clippingRectangle: nil.
! !

!ImageEditView methodsFor:'edit modes'!

editMode

    editMode isNil ifTrue: [editMode := 'point'].
    ^editMode
!

editMode:aMode

    editMode := aMode
!

mouseKeyColorMode

    ^mouseKeyColorMode printString
!

mouseKeyColorMode:aMode

    mouseKeyColorMode := aMode asInteger
! !

!ImageEditView methodsFor:'event handling'!

buttonMotion:state x:x y:y

    self drawCursorAt: x@y.
    self selectedColor notNil & image notNil & (self imageContainsPoint: x@y) & (editMode = 'point')
        ifTrue: [^self pointAt: x@y].
!

buttonPress:button x:x y:y

    self drawCursorAt: x@y.
    self selectedColor notNil & image notNil & (self imageContainsPoint: x@y)
    ifTrue:
    [    
        undoImage := image copy.
        mouseKeyColorMode := button.
        self perform: (editMode, 'At:') asSymbol with: x@y.
    ].
!

buttonRelease:state x:x y:y

    Cursor normal show
! !

!ImageEditView methodsFor:'image drawing'!

boxAt: aPoint

    |choosedBox|
    choosedBox := self getRectangleFrom: aPoint emphasis: #box.
    image rectangle: (choosedBox origin//magnification extent: (choosedBox extent//magnification)) withColor: self selectedColor.
    image restored.
    self redraw: (choosedBox expandedBy: 1@1).    
    modified := true.
!

copyAt: aPoint

    |choosedBox|
    choosedBox := self getRectangleFrom: aPoint emphasis: #inverseFilledBox.
    Clipboard := image subImageIn: (choosedBox origin//magnification extent: (choosedBox extent//magnification)).
    self redraw: choosedBox    

!

fillAt: aPoint

    Cursor execute showWhile:
    [
        |filledPoints|
        filledPoints := image floodFillAt: aPoint//magnification withColor: self selectedColor.
        image restored.
        filledPoints size < 300
        ifTrue:
        [
            filledPoints do: [:p| Delay waitForMilliseconds:0.7. self redraw: ((p * magnification extent: magnification) expandedBy: 1@1)]
        ]
        ifFalse:
        [
            self invalidate
        ].
        modified := true
    ]
!

filledBoxAt: aPoint

    |choosedBox|
    choosedBox := self getRectangleFrom: aPoint emphasis: #filledBox.
    image fillRectangle: (choosedBox origin//magnification extent: (choosedBox extent//magnification)) withColor: self selectedColor.
    image restored.
    self redraw: (choosedBox expandedBy: 1@1).    
    modified := true.
!

getRectangleFrom: aPoint emphasis: emphasis

    |currentPoint currentExtent firstPoint lastCurrentPoint gridCorrection p|
    firstPoint := currentPoint := lastCurrentPoint := aPoint//magnification*magnification.
    self isGridOn ifFalse: [gridCorrection := 0] ifTrue: [gridCorrection := 1].
    [Display anyButtonPressed]
    whileTrue:
    [   
        currentPoint := (0@0) max: (image extent * magnification min: (p := self translation negated + (device translatePoint: self sensor mousePoint from:device rootView id to:self id))).
        currentPoint := currentPoint//magnification*magnification.
        currentExtent := (firstPoint - currentPoint) abs.
        self drawCursorAt: p.

        currentPoint ~= lastCurrentPoint ifTrue:
        [          
            self redraw: ((firstPoint min: lastCurrentPoint) - 1 extent: (firstPoint - lastCurrentPoint) abs + 2).
            emphasis = #inverseFilledBox
            ifTrue:
            [
                self xoring: [self fillRectangle: ((firstPoint min: currentPoint) + 1 extent: currentExtent - gridCorrection)]
            ].
            emphasis = #box
            ifTrue:
            [
                self selectedColor ~= Color noColor
                    ifTrue: [self paint: self selectedColor]
                    ifFalse: [self paint: self viewBackground].
                self lineWidth: magnification x.
                self displayRectangle: ((firstPoint min: currentPoint) + (magnification x//2) + 1 extent: (currentExtent - (magnification x)) - gridCorrection + 1).
                self lineWidth: 1.
            ].
            emphasis = #filledBox
            ifTrue:
            [
                self selectedColor ~= Color noColor
                    ifTrue: [self paint: self selectedColor]
                    ifFalse: [self paint: self viewBackground].
                self fillRectangle: ((firstPoint min: currentPoint) + 1 extent: currentExtent - gridCorrection).
            ].
        ]. 
        self drawLabel: currentPoint//magnification.
        lastCurrentPoint := currentPoint.
    ].
    Cursor normal show.
    ^(firstPoint min: currentPoint) extent: (firstPoint - currentPoint) abs
!

pasteAt: aPoint

    Object errorSignal handle:
    [:ex|
        self warn: 'Pasting at selected point failed!!'.
    ] 
    do:
    [   
        |imagePoint|
        imagePoint := aPoint//magnification.
        image copyFrom: Clipboard x:0 y:0 toX: imagePoint x y: imagePoint y width: Clipboard width height: Clipboard height.
        image restored.
        self redraw: ((imagePoint * magnification extent: (Clipboard extent * magnification)) expandedBy: 1@1).
        self drawLabel: imagePoint.
        modified := true.
    ]
!

pointAt: aPoint

    |imagePoint|
    imagePoint := aPoint//magnification.
    image atImageAndMask: imagePoint put: self selectedColor.
    image restored.
    self redraw: ((imagePoint * magnification extent: magnification) expandedBy: 1@1).
    self drawLabel: imagePoint.
    modified := true.

!

undo

    undoImage notNil
    ifTrue:
    [
        modified := false.
        self image: undoImage.
        self invalidate
    ]
! !

!ImageEditView methodsFor:'image editing'!

flipHorizontal

    self image: image flipHorizontal

!

flipVertical

    self image: image flipVertical

!

negativeImage

    self image: image negative

!

resizeImage

    |b newSize|

    b := EnterBox new.
    b title:'Image new size:'.
    b okText:'OK'.
    b abortText:'Cancel'.
    b initialText:image extent printString.
    b showAtPointer.
    (b accepted and: [(newSize := Object readFromString: b contents onError:nil) notNil])
    ifTrue:
    [
        self image: (image magnifiedBy: newSize/image extent)
    ].
!

rotateImage

    |b rotation|

    b := EnterBox new.
    b title:'Image new rotation:'.
    b okText:'OK'.
    b abortText:'Cancel'.
    b initialText: '0'.
    b showAtPointer.
    (b accepted and: [(rotation := Object readFromString: b contents onError:nil) notNil])
    ifTrue:
    [   Object errorSignal handle:
        [:ex|
            self warn: 'Image rotation failed.\' withCRs, 'An increase of image depth could help.'.
        ] 
        do:
        [   
            self image: (image hardRotated: rotation)
        ]
    ].
! !

!ImageEditView methodsFor:'image emphasis'!

drawCursorAt: aPoint

    image isNil ifTrue: [^Cursor stop show].
    ((0@0 extent: image extent * magnification) containsPoint: aPoint)
         ifFalse:[Cursor stop show]
         ifTrue: [Cursor crossHair show].

!

drawFrame

    self paint:Color black.
    self lineWidth: (magnification x//3 min: 3).
    self displayRectangle: ((0@0) extent:(image extent * magnification) + 2).
    self lineWidth:1.
!

drawFrameAt: aPoint

    self isGridOn
    ifTrue:
    [   
        |lineStartingPoint lineEndingPoint oldColor|
        lineStartingPoint := aPoint//magnification*magnification.
        lineEndingPoint   := aPoint//magnification*magnification + magnification.
        oldColor := self paint.
        self xoring:
        [
            self displayLineFrom: lineEndingPoint 
                              to: (lineEndingPoint x)@(lineStartingPoint y).
            self displayLineFrom: lineEndingPoint 
                              to: (lineStartingPoint x)@(lineEndingPoint y).
        ].
        self paint: oldColor.
    ]
!

drawLabel: aLabel

    coordInfoBlock notNil
    ifTrue:
    [         
        coordInfoBlock value: aLabel printString
    ]
!

drawMaskPointAt: aPoint

    |sizeOfMaskPoint|
    sizeOfMaskPoint := magnification//3.
    self xoring: [self fillRectangle: (aPoint + sizeOfMaskPoint extent: sizeOfMaskPoint)].
   
! !

!ImageEditView methodsFor:'image setting'!

image:anImage

    (anImage isImage and: [image isNil or: [self checkModified]])
    ifTrue:
    [
        super image: anImage.
        ^self
    ].
    ^nil
!

loadFromFile: aFileName

    Object errorSignal handle:
    [:exeption|
        self warn: exeption errorString.
        ^nil
    ] 
    do:
    [  
        (self image: (Image fromFile: aFileName)) notNil
        ifTrue:
        [
            imageReaderClass := ImageReader allSubclasses
                detect: [:cls| cls isValidImageFile: aFileName]
                ifNone: [self error: 'Unknown image file format!!']
        ]
        ifFalse:
        [
            self error: 'Image on file not found!!'
        ]
    ] 
!

loadFromMessage: aMessage

    (self resourceMessage: aMessage) isNil
    ifTrue:
    [
        ^nil
    ].

    ^self loadfromClass: resourceClass andSelector: resourceSelector

   
!

loadfromClass: aClassOrSymbol andSelector: aStringOrSymbol

    |aClass|       
    imageReaderClass := nil.
    self resourceClass: aClassOrSymbol.
    self resourceSelector: aStringOrSymbol.
    aClass := Smalltalk at: resourceClass. 
    (aClass isClass and: [aClass class implements: resourceSelector])
    ifTrue:
    [ 
        ^self image: (aClass perform: resourceSelector) copy
    ].
    self warn: 'No class or selector for evaluation detected!!'.
    ^nil
! !

!ImageEditView methodsFor:'initialization'!

initialize

    super initialize.

    magnification := 1@1.
    gridMagnification := 8@8.
    modified := false.
    mouseKeyColorMode := 1.
    resourceClass := resourceSelector := ''.
    selectColors := Array with: nil with: nil.
! !

!ImageEditView methodsFor:'menu actions'!

changeMagnification

    |newMag|
    EnterBox new
        title:'Magnification (magX @ magY)';
        okText:'Apply';
        abortText:'Abort';
        action:[:string | newMag := (Object readFromString:string onError:nil)];
        initialText:(magnification printString);
        showAtPointer.

    newMag notNil ifTrue:[
        self magnification:newMag.
    ].
!

loadFromClass

    self loadFromMessage: (ResourceSelectionBrowser
        openOnSuperclass: ApplicationModel
        class: self resourceClass
        selector: self resourceSelector
        resourceTypes: #(image fileImage))
        

!

print
    |stream psgc|

    image isNil ifTrue: [^nil].

    Printer supportsPostscript ifFalse:[
        self warn:'need a postscript printer'.
        ^ self
    ].

    stream := Printer newNative.
    stream isNil ifTrue:[
        self warn:'cannot open printer stream'.
        ^ nil
    ].

    self withWaitCursorDo:[
        psgc := PSGraphicsContext on:stream. "/  extent:(1.0 @ 1.0).
        psgc displayForm: (image magnifiedBy: magnification) x:0 y:0.
        psgc close.
    ]
!

save

    self saveImageOrMask: #image.

!

saveAs
    "save contents into a file 
     - ask user for filename using a fileSelectionBox."

    self saveImageFileAs
!

saveImageFileAs

    self saveImageFileAs:
        (FileSelectionBrowser
            title: 'Saving Image As'
            requestFileName: self image fileName
            fileFilters: FileSelectionBrowser saveImageFileNameFilters)
!

saveImageFileAs: aFileName

    aFileName isNil ifTrue: [^nil].
    image notNil & (aFileName readStream nextWord size > 0)
    ifTrue:
    [
        image fileName: aFileName.
        self saveImageOrMask: #image
    ]
    ifFalse:
    [
        self warn: 'No image or file name for saving detected!!'
    ]
!

saveImageMaskFileAs

    self saveImageMaskFileAs:
        (FileSelectionBrowser 
            requestFileName: self image fileName 
            fileFilters: FileSelectionBrowser saveImageFileNameFilters)
!

saveImageMaskFileAs: aFileName

    aFileName isNil ifTrue: [^nil].
    (image notNil and: [image mask notNil])
    ifTrue:
    [
        image mask fileName: aFileName.
        self saveImageOrMask: #mask
    ]
    ifFalse:
    [
        self warn: 'No image or image mask detected!!'
    ]
!

saveImageOrMask: what

    Object errorSignal handle:
    [:ex|
        self warn: ex errorString.
        ^nil                                 
    ] 
    do:
    [   
        |fileName|
        image isNil ifTrue: [^self error: 'No image to save!!'].
        what = #image
        ifTrue:
        [   
            image fileName isNil ifTrue: [^self error: 'No file name for image detected!!'].
            fileName := image fileName asFilename
        ].
        what = #mask
        ifTrue:
        [   
            image mask isNil ifTrue: [^self error: 'No image mask to save!!'].
            image mask fileName isNil ifTrue: [^self error: 'No file name for image mask detected!!'].
            fileName := image mask fileName asFilename
        ].
        (fileName suffix = 'tiff') | (fileName suffix = 'tif') ifTrue: [imageReaderClass := TIFFReader].
        fileName suffix = 'xpm' ifTrue: [imageReaderClass := XPMReader].
        fileName suffix = 'xbm' ifTrue: [imageReaderClass := XBMReader].
        fileName suffix = 'gif' ifTrue: [self halt. imageReaderClass := GIFReader].
        (fileName suffix = 'jpg') | (fileName suffix = 'jpeg') ifTrue: [imageReaderClass := JPEGReader].
        imageReaderClass isNil ifTrue: [imageReaderClass := XPMReader. image fileName: image fileName, '.xpm'].
        what = #image ifTrue: [image saveOn: image fileName using: imageReaderClass. modified := false].
        what = #mask ifTrue: [image mask saveOn: image mask fileName using: imageReaderClass].
    ]

       
!

saveMethod

    Object errorSignal handle:
    [:ex|
        self warn: ex errorString.
        ^nil                                 
    ] 
    do:
    [   
        |imageStoreStream cls|  
        (self resourceSelector trimBlanks size = 0) | (cls := Smalltalk at: self resourceClass) isClass not 
            ifTrue: [^self error: 'No class and selector for image detected!!'].
        self image storeOn: (imageStoreStream := WriteStream on: '').
        ByteCodeCompiler 
            compile: (self resourceSelector,
                '\' withCRs, '    "ImageEditor openOnClass:self andSelector:#', self resourceSelector, '"',
                '\\' withCRs, '    <resource: #image>\' withCRs, '    ^', imageStoreStream contents)
            forClass: cls class inCategory: 'resources'.
        modified := false.
    ]
!

saveMethodAs

    (self resourceMessage: (ResourceSelectionBrowser
            openOnSuperclass: ApplicationModel
            class: self resourceClass
            selector: self resourceSelector
            resourceTypes: #(image fileImage))) notNil
    ifTrue:
    [
        ^self saveMethod
    ].
    ^nil
! !

!ImageEditView methodsFor:'queries'!

heightOfContents
    "return the images height"

    image isNil ifTrue:[^ 0].
    ^ (image height * magnification y) rounded
!

imageContainsPoint: aPoint

    ^image notNil and:
        [((0@0 corner:(image extent) - 1) containsPoint: (((aPoint - margin + 1) / magnification) floor))]
!

widthOfContents
    "return the images width"

    image isNil ifTrue:[^ 0].
    ^ (image width * magnification x) rounded
! !

!ImageEditView methodsFor:'release'!

destroy

    undoImage := nil.
    Clipboard := nil.
    super destroy.

! !

!ImageEditView methodsFor:'testing'!

checkModified

    modified
    ifTrue:
    [
        ((YesNoBox title:'Image was modified!!')        
            noText:'Cancel';
            yesText:'Waste it and proceed';
            showAtPointer;
            accepted) ifFalse: [^false].
        modified := false
    ].
    ^true
!

isGridOn

    ^magnification > gridMagnification
! !

!ImageEditView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/ImageEditView.st,v 1.52 1997-12-11 12:43:42 tz Exp $'
! !
